# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import datetime
import os
import sys

sys.path.append(os.path.abspath("./_pygments"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "autoray"
copyright = f"2019-{datetime.date.today().year}, Johnnie Gray"
author = "Johnnie Gray"

try:
    from autoray import __version__

    release = __version__
except ImportError:
    try:
        from importlib.metadata import version as _version

        release = _version("autoray")
    except ImportError:
        release = "0.0.0+unknown"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_nb",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.napoleon",
    "sphinx.ext.linkcode",
    "sphinx_copybutton",
    "autoapi.extension",
]

nb_execution_mode = "off"
myst_heading_anchors = 4
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
]

# sphinx-autoapi
autoapi_dirs = ["../autoray"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"

html_theme_options = {
    "sidebar_hide_name": True,
    "light_css_variables": {
        #     "color-brand-primary": "hsl(72, 75%, 40%)",
        #     "color-brand-content": "hsl(238, 50%, 60%)",
        "font-stack": "Atkinson Hyperlegible, sans-serif",
        "font-stack--monospace": "Spline Sans Mono, monospace",
        "font-stack--headings": "Spline Sans Mono, monospace",
    },
    # "dark_css_variables": {
    #     "color-brand-primary": "hsl(72, 75%, 60%)",
    #     "color-brand-content": "hsl(238, 75%, 70%)",
    # },
    "light_logo": "autoray-header.png",
    "dark_logo": "autoray-header.png",
}

html_css_files = ["my-styles.css"]
html_static_path = ["_static"]
html_favicon = "_static/autoray.ico"

pygments_style = "_pygments_light.MarianaLight"
pygments_dark_style = "_pygments_dark.MarianaDark"


def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    import inspect

    import autoray

    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except OSError:
        lineno = None

    if lineno:
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname(autoray.__file__))

    if "+" in autoray.__version__:
        return (
            f"https://github.com/jcmgray/autoray/blob/"
            f"HEAD/autoray/{fn}{linespec}"
        )
    else:
        return (
            f"https://github.com/jcmgray/autoray/blob/"
            f"v{autoray.__version__}/autoray/{fn}{linespec}"
        )



import pytest

import autoray as ar

from .test_autoray import BACKENDS


@pytest.mark.parametrize(
    "backend",
    [
        b
        for b in BACKENDS
        if any(
            b.values[0] == other
            for other in (
                "numpy",
                "jax",
                "torch",
                "cupy",
                "dask",
            )
        )
    ],
)
@pytest.mark.parametrize(
    "dist,args,kwargs",
    [
        ("binomial", (7, 0.424), {"size": (3, 4)}),
        ("choice", ([11.1 * i for i in range(100)],), {"size": (3, 4)}),
        ("choice", ([11.1 * i for i in range(1000)],), {}),
        ("exponential", (), {"size": (3, 4)}),
        ("exponential", (), {}),
        ("gumbel", (), {"size": (3, 4)}),
        ("gumbel", (), {}),
        ("integers", (100, 1000), {"size": (3, 4)}),
        ("integers", (100, 1000), {}),
        ("normal", (), {"size": (3, 4)}),
        ("normal", (), {}),
        ("permutation", ([11.1 * i for i in range(100)],), {}),
        ("poisson", (100,), {"size": (3, 4, 5)}),
        ("random", (), {"size": (3, 4)}),
        ("random", (), {}),
        ("uniform", (), {"size": (3, 4)}),
        ("uniform", (), {}),
    ],
)
def test_random_default_rng(backend, dist, args, kwargs):
    if dist in ("choice", "permutation"):
        args = (ar.do("array", args[0], like=backend), *args[1:])

    if dist == "permutation" and backend == "dask":
        pytest.xfail("bug: https://github.com/dask/dask/issues/12029")

    if backend == "torch" and dist in (
        "binomial",
        "choice",
        "exponential",
        "gumbel",
        "poisson",
    ):
        pytest.xfail(f"torch: no {dist} interface yet.")

    if backend == "cupy" and dist in (
        "choice",
        "gumbel",
        "normal",
        "permutation",
    ):
        pytest.xfail(f"torch: no {dist} interface yet.")

    seed = 42
    seed2 = 43

    rng = ar.do("random.default_rng", seed, like=backend)
    x = ar.do("to_numpy", getattr(rng, dist)(*args, **kwargs))
    if "size" in kwargs:
        assert ar.do("shape", x) == kwargs["size"]
    y = ar.do("to_numpy", getattr(rng, dist)(*args, **kwargs))
    assert not ar.do("allclose", x, y)
    rng = ar.do("random.default_rng", seed2, like=backend)
    z = ar.do("to_numpy", getattr(rng, dist)(*args, **kwargs))
    assert not ar.do("allclose", x, z)
    rng = ar.do("random.default_rng", seed, like=backend)
    x2 = ar.do("to_numpy", getattr(rng, dist)(*args, **kwargs))
    assert ar.do("allclose", x, x2)


def test_jax_jit_random():
    pytest.importorskip("jax")

    @ar.autojit(backend="jax")
    def f(seed):
        rng = ar.do("random.default_rng", seed)
        return rng.normal(size=(3, 4))

    x1 = ar.do("to_numpy", f(ar.do("array", 42)))
    x2 = ar.do("to_numpy", f(ar.do("array", 42)))
    assert ar.do("allclose", x1, x2)
    x3 = ar.do("to_numpy", f(ar.do("array", 43)))
    assert not ar.do("allclose", x1, x3)



try:
    # -- Distribution mode --
    # import from _version.py generated by setuptools_scm during release
    from ._version import version as __version__
except ImportError:
    # -- Source mode --
    try:
        # use setuptools_scm to get the current version from src using git
        from pathlib import Path as _Path

        from setuptools_scm import get_version as _gv

        __version__ = _gv(_Path(__file__).parent.parent)
    except ImportError:
        # setuptools_scm is not available, use a default version
        __version__ = "0.0.0+unknown"


# useful constants
from math import (
    e,
    inf,
    nan,
    pi,
)

from . import lazy
from .autoray import (
    astype,
    backend_like,
    compose,
    conj,
    dag,
    do,
    get_backend,
    get_common_dtype,
    get_dtype_name,
    get_lib_fn,
    get_namespace,
    imag,
    infer_backend,
    infer_backend_multi,
    is_array,
    ndim,
    # the numpy mimic submodule
    numpy,
    real,
    register_backend,
    register_function,
    reshape,
    set_backend,
    shape,
    size,
    to_backend_dtype,
    to_numpy,
    transpose,
    # tree utilities
    tree_apply,
    tree_flatten,
    tree_iter,
    tree_map,
    tree_unflatten,
)
from .compiler import autojit

__all__ = (
    "do",
    "get_backend",
    "set_backend",
    "backend_like",
    "infer_backend",
    "infer_backend_multi",
    "get_lib_fn",
    "get_namespace",
    "shape",
    "ndim",
    "size",
    "conj",
    "transpose",
    "dag",
    "real",
    "imag",
    "reshape",
    "to_backend_dtype",
    "get_dtype_name",
    "get_common_dtype",
    "astype",
    "to_numpy",
    "register_backend",
    "register_function",
    # tree utilities
    "is_array",
    "tree_map",
    "tree_iter",
    "tree_apply",
    "tree_flatten",
    "tree_unflatten",
    "compose",
    # the numpy mimic submodule
    "numpy",
    # abstract function compilation
    "autojit",
    # lazy array library
    "lazy",
    # math constants,
    "e",
    "pi",
    "inf",
    "nan",
)



from . import linalg
from .core import (
    Function,
    LazyArray,
    Variable,
    add,
    angle,
    arccos,
    arccosh,
    arcsin,
    arcsinh,
    arctan,
    arctanh,
    argsort,
    array,
    ascend,
    broadcast_to,
    clip,
    compute,
    concatenate,
    conj,
    cos,
    cosh,
    descend,
    diag,
    diagonal,
    einsum,
    empty,
    equal,
    exp,
    expand_dims,
    eye,
    flip,
    floordivide,
    get_source,
    identity,
    imag,
    kron,
    log,
    log2,
    log10,
    matmul,
    multiply,
    ones,
    permute_dims,
    prod,
    real,
    reshape,
    shared_intermediates,
    sign,
    sin,
    sinh,
    sort,
    split,
    sqrt,
    stack,
    take,
    take_along_axis,
    tan,
    tanh,
    tensordot,
    trace,
    transpose,
    truedivide,
    where,
    zeros,
)
from .core import abs_ as abs
from .core import complex_ as complex
from .core import max_ as max
from .core import min_ as min
from .core import sum_ as sum

__all__ = (
    "abs",
    "add",
    "angle",
    "arccos",
    "arccosh",
    "arcsin",
    "arcsinh",
    "arctan",
    "arctanh",
    "argsort",
    "array",
    "ascend",
    "broadcast_to",
    "clip",
    "complex",
    "compute",
    "concatenate",
    "conj",
    "conj",
    "cos",
    "cosh",
    "descend",
    "diag",
    "diagonal",
    "einsum",
    "empty",
    "equal",
    "exp",
    "expand_dims",
    "eye",
    "flip",
    "floordivide",
    "Function",
    "get_source",
    "identity",
    "imag",
    "kron",
    "LazyArray",
    "linalg",
    "log",
    "log10",
    "log2",
    "matmul",
    "max",
    "min",
    "multiply",
    "ones",
    "permute_dims",
    "prod",
    "real",
    "reshape",
    "shared_intermediates",
    "sign",
    "sin",
    "sinh",
    "sort",
    "split",
    "sqrt",
    "stack",
    "sum",
    "take",
    "take_along_axis",
    "tan",
    "tanh",
    "tensordot",
    "trace",
    "transpose",
    "truedivide",
    "Variable",
    "where",
    "zeros",
)


try:
    from opt_einsum.backends.dispatch import _aliases

    _aliases["autoray"] = "autoray.lazy"
except ImportError:  # pragma: no cover
    pass



import pytest
from numpy.testing import assert_allclose

from autoray import autojit, do, infer_backend, shape, to_numpy

from .test_autoray import BACKENDS, gen_rand

BACKENDS = [
    p for p in BACKENDS if p.values[0] in ("jax", "torch", "tensorflow")
]


def modified_gram_schmidt(X):
    Q = []
    for j in range(0, shape(X)[0]):
        q = X[j, :]
        for i in range(0, j):
            rij = do("tensordot", do("conj", Q[i]), q, axes=1)
            q = q - rij * Q[i]
        rjj = do("linalg.norm", q, 2)
        Q.append(q / rjj)
    return do("stack", tuple(Q), axis=0)


@pytest.fixture
def mgs_case():
    x = gen_rand((10, 10), "numpy")
    y = modified_gram_schmidt(x)
    return x, y


@pytest.mark.parametrize("share_intermediates", [False, True])
@pytest.mark.parametrize("nested", [False, True])
def test_compile_python(mgs_case, share_intermediates, nested):
    x, y = mgs_case
    compiler_opts = {"python": {"share_intermediates": share_intermediates}}
    mgs = autojit(modified_gram_schmidt, compiler_opts=compiler_opts)
    if nested:
        mgs = autojit(mgs, compiler_opts=compiler_opts)
    y2 = mgs(x)
    assert_allclose(y, y2)


@pytest.mark.parametrize("backend", BACKENDS)
def test_others_numpy(backend, mgs_case):
    x, y = mgs_case
    mgs = autojit(modified_gram_schmidt)
    y2 = mgs(x, backend=backend)
    assert infer_backend(y2) == "numpy"
    assert_allclose(y, y2)


@pytest.mark.parametrize("backend", BACKENDS)
def test_autodispatch(backend, mgs_case):
    x, y = mgs_case
    x = do("array", x, like=backend)
    mgs = autojit(modified_gram_schmidt)
    y2 = mgs(x, backend=backend)
    assert infer_backend(y2) == backend
    assert_allclose(y, to_numpy(y2))


def test_complicated_signature():
    @autojit
    def foo(a, b, c):
        a1, a2 = a
        b1 = b["1"]
        c1, c2 = c["sub"]
        return do("sum", do("stack", (a1, a2, b1, c1, c2)), axis=0)

    x = do("random.uniform", size=(5, 7), like="numpy")
    y = foo((x[0, :], x[1, :]), {"1": x[2, :]}, c={"sub": (x[3, :], x[4, :])})
    assert_allclose(y, x.sum(0))


def test_multi_output():
    @autojit
    def foo(a, b, c):
        a = a - do("sum", b)
        b = b - do("sum", a)
        return a + c, b - c

    a = gen_rand((2, 3), "numpy")
    b = gen_rand((4, 5), "numpy")
    x, y = foo(a, b, 1)

    assert_allclose(x, a - b.sum() + 1)
    assert_allclose(y, b - (a - b.sum()).sum() - 1)
